@@ -1,4 +1,4 @@
-/* Copyright 2007 Matthew Barnes
+/* Copyright 2007, 2008 Matthew Barnes
  *
  * This file is part of GNOME Video Arcade.
  *
@@ -28,7 +28,9 @@
 
 #include "gva-error.h"
 #include "gva-mame-process.h"
+#include "gva-mute-button.h"
 #include "gva-preferences.h"
+#include "gva-ui.h"
 
 /*****************************************************************************
  * Private utilities for MAME backends
@@ -110,6 +112,7 @@ gva_mame_command (const gchar *arguments,
         if (process->error != NULL)
         {
                 g_propagate_error (error, process->error);
+                process->error = NULL;
                 status = -1;
                 goto fail;
         }
@@ -173,7 +176,7 @@ gva_mame_get_config_value (const gchar *config_key,
          * ...
          */
 
-        num_lines = g_strv_length (lines);
+        num_lines = (lines != NULL) ? g_strv_length (lines) : 0;
 
         for (ii = 0; ii < num_lines; ii++)
         {
@@ -241,17 +244,11 @@ gva_mame_has_config_value (const gchar *config_key)
         GError *error = NULL;
 
         config_value = gva_mame_get_config_value (config_key, &error);
+        result = (config_value != NULL && *config_value != '\0');
+        g_free (config_value);
 
-        if (config_value != NULL)
+        if (error != NULL)
         {
-                result = TRUE;
-                g_free (config_value);
-        }
-        else
-        {
-                result = FALSE;
-                g_assert (error != NULL);
-
                 /* Suppress warnings about unknown configuration
                  * keys, since that's what we're testing for. */
                 if (error->code == GVA_ERROR_MAME)
@@ -263,6 +260,33 @@ gva_mame_has_config_value (const gchar *config_key)
         return result;
 }
 
+/**
+ * gva_mame_get_search_paths:
+ * @config_key: a configuration key
+ * @error: return location for a #GError, or %NULL
+ *
+ * Returns the value of @config_key as an ordered list of search paths.
+ * This is only appropriate for configuration keys containing a list of
+ * search paths, such as "rompath" or "samplepath".  If an error occurs,
+ * or if @config_key is not found in MAME's configuration, the function
+ * returns %NULL and sets @error.
+ *
+ * Returns: a newly-allocated %NULL-terminated array of paths, or %NULL.
+ *          Use g_strfreev() to free it.
+ **/
+gchar **
+gva_mame_get_search_paths (const gchar *config_key,
+                           GError **error)
+{
+        gchar *config_value;
+
+        config_value = gva_mame_get_config_value (config_key, error);
+        if (config_value == NULL)
+                return NULL;
+
+        return g_strsplit (config_value, gva_mame_get_path_sep (), -1);
+}
+
 /**
  * gva_mame_get_input_files:
  * @error: return location for a @GError, or %NULL
@@ -276,16 +300,16 @@ gva_mame_has_config_value (const gchar *config_key)
 GHashTable *
 gva_mame_get_input_files (GError **error)
 {
-        GHashTable *hash_table = NULL;
+        GHashTable *hash_table;
         const gchar *basename;
-        gchar *inppath;
+        const gchar *directory;
         GDir *dir;
 
-        inppath = gva_mame_get_config_value ("input_directory", error);
-        dir = (inppath != NULL) ? g_dir_open (inppath, 0, error) : NULL;
+        directory = gva_mame_get_input_directory (error);
 
+        dir = g_dir_open (directory, 0, error);
         if (dir == NULL)
-                goto exit;
+                return NULL;
 
         hash_table = g_hash_table_new_full (
                 g_str_hash, g_str_equal,
@@ -298,7 +322,7 @@ gva_mame_get_input_files (GError **error)
                 GIOChannel *channel;
                 GError *local_error = NULL;
 
-                filename = g_build_filename (inppath, basename, NULL);
+                filename = g_build_filename (directory, basename, NULL);
 
                 channel = g_io_channel_new_file (filename, "r", &local_error);
 
@@ -329,9 +353,6 @@ gva_mame_get_input_files (GError **error)
 
         g_dir_close (dir);
 
-exit:
-        g_free (inppath);
-
         return hash_table;
 }
 
@@ -361,6 +382,100 @@ gva_mame_list_xml (GError **error)
 
 /**
  * gva_mame_verify_roms:
+ * @name: the name of a ROM set
+ * @error: return location for a #GError, or %NULL
+ *
+ * Verifies the contents of the ROM set @name and returns the status, which
+ * may be "good", "bad", "best available", "not found", or "not supported".
+ * If an error occurs, it returns %NULL and sets @error.
+ *
+ * Returns: verification status, or %NULL
+ **/
+gchar *
+gva_mame_verify_roms (const gchar *name,
+                      GError **error)
+{
+        gchar *arguments;
+        gchar **stdout_lines;
+        gchar **stderr_lines;
+        gchar *status = NULL;
+        gint exit_status;
+        gint ii;
+
+        g_return_val_if_fail (name != NULL, NULL);
+
+        /* Execute the command "${mame} -verifyroms %{name}". */
+        arguments = g_strdup_printf ("-verifyroms %s", name);
+        exit_status = gva_mame_command (
+                arguments, &stdout_lines, &stderr_lines, error);
+        g_free (arguments);
+
+        if (exit_status < 0)
+                return NULL;
+
+        /* First try to extract a status from standard output. */
+        for (ii = 0; status == NULL && stdout_lines[ii] != NULL; ii++)
+                gva_mame_verify_parse (stdout_lines[ii], NULL, &status);
+
+        /* If that fails, try to extract a status from standard error. */
+        for (ii = 0; status == NULL && stderr_lines[ii] != NULL; ii++)
+                gva_mame_verify_parse (stderr_lines[ii], NULL, &status);
+
+        g_strfreev (stdout_lines);
+        g_strfreev (stderr_lines);
+
+        return status;
+}
+
+/**
+ * gva_mame_verify_samples:
+ * @name: the name of a sample set
+ * @error: return location for a #GError, or %NULL
+ *
+ * Verifies the contents of the sample set @name and returns the status,
+ * which may be "good", "bad", "best available", "not found", or "not
+ * supported".  If an error occurs, it returns %NULL and sets @error.
+ *
+ * Returns: verification status, or %NULL
+ **/
+gchar *
+gva_mame_verify_samples (const gchar *name,
+                         GError **error)
+{
+        gchar *arguments;
+        gchar **stdout_lines;
+        gchar **stderr_lines;
+        gchar *status = NULL;
+        gint exit_status;
+        gint ii;
+
+        g_return_val_if_fail (name != NULL, NULL);
+
+        /* Execute the command "${mame} -verifysamples %{name}". */
+        arguments = g_strdup_printf ("-verifysamples %s", name);
+        exit_status = gva_mame_command (
+                arguments, &stdout_lines, &stderr_lines, error);
+        g_free (arguments);
+
+        if (exit_status < 0)
+                return NULL;
+
+        /* First try to extract a status from standard output. */
+        for (ii = 0; status == NULL && stdout_lines[ii] != NULL; ii++)
+                gva_mame_verify_parse (stdout_lines[ii], NULL, &status);
+
+        /* If that fails, try to extract a status from standard error. */
+        for (ii = 0; status == NULL && stderr_lines[ii] != NULL; ii++)
+                gva_mame_verify_parse (stderr_lines[ii], NULL, &status);
+
+        g_strfreev (stdout_lines);
+        g_strfreev (stderr_lines);
+
+        return status;
+}
+
+/**
+ * gva_mame_verify_all_roms:
  * @error: return location for a #GError, or %NULL
  *
  * Spawns a "MAME -verifyroms" child process and returns a #GvaProcess so
@@ -370,7 +485,7 @@ gva_mame_list_xml (GError **error)
  * Returns: a new #GvaProcess, or %NULL
  **/
 GvaProcess *
-gva_mame_verify_roms (GError **error)
+gva_mame_verify_all_roms (GError **error)
 {
         /* Execute the command "${mame} -verifyroms". */
         return gva_mame_process_spawn (
@@ -378,7 +493,7 @@ gva_mame_verify_roms (GError **error)
 }
 
 /**
- * gva_mame_verify_samples:
+ * gva_mame_verify_all_samples:
  * @error: return location for a #GError, or %NULL
  *
  * Spawns a "MAME -verifysamples" child process and returns a #GvaProcess so
@@ -388,13 +503,109 @@ gva_mame_verify_roms (GError **error)
  * Returns: a new #GvaProcess, or %NULL
  **/
 GvaProcess *
-gva_mame_verify_samples (GError **error)
+gva_mame_verify_all_samples (GError **error)
 {
         /* Execute the command "${mame} -verifysamples". */
         return gva_mame_process_spawn (
                 "-verifysamples", G_PRIORITY_DEFAULT_IDLE, error);
 }
 
+/**
+ * gva_mame_verify_parse:
+ * @line: output line from a MAME process
+ * @out_name: return location for the game name, or %NULL
+ * @out_status: return location for the status, or %NULL
+ *
+ * Attempts to extract a status from a line of output generated by a
+ * "MAME -verifyroms" or "MAME -verifysamples" child process.  If successful,
+ * it sets @out_name to a newly-allocated string containing the game name and
+ * @out_status to a newly-allocated string containing the verification status
+ * and returns %TRUE.
+ *
+ * Known status values are "good", "bad", "best available", "not found" and
+ * "not supported".
+ *
+ * Returns: %TRUE if the parse was successful, %FALSE otherwise
+ **/
+gboolean
+gva_mame_verify_parse (const gchar *line,
+                       gchar **out_name,
+                       gchar **out_status)
+{
+        gchar *copy;
+        const gchar *name = NULL;
+        const gchar *sep = " \"[]:!";
+        const gchar *status = NULL;
+        const gchar *token;
+
+        g_return_val_if_fail (line != NULL, FALSE);
+
+        /* Output for -verifyroms is as follows:
+         *
+         * romset puckman is good
+         * romset puckmana [puckman] is good
+         * romset puckmanf [puckman] is good
+         * ...
+         * digdug  : 51xx.bin (1024 bytes) - NOT FOUND
+         * digdug  : 53xx.bin (1024 bytes) - NOT FOUND
+         * romset digdug is bad
+         * ...
+         *
+         * - Status may be "good", "bad", or "best available".
+         *
+         * - When checking individual sets, status may also be
+         *   "not found" or "not supported".
+         *
+         * - Older MAMEs used "correct" and "incorrect" instead of
+         *   "good" and "bad".  Convert to the newer form if seen.
+         *
+         * - Similar output for -verifysamples.
+         */
+
+        copy = g_strchomp (g_ascii_strdown (line, -1));
+
+        if ((token = strtok (copy, sep)) == NULL)
+                goto exit;
+
+        if (strcmp (token, "error") == 0)
+                if ((token = strtok (NULL, sep)) == NULL)
+                        goto exit;
+
+        if (strcmp (token, "romset") != 0)
+                if (strcmp (token, "sampleset") != 0)
+                        goto exit;
+
+        name = strtok (NULL, sep);
+        while ((token = strtok (NULL, sep)) != NULL)
+                status = token;
+
+        if (name == NULL || status == NULL)
+                goto exit;
+
+        /* Normalize the status. */
+        if (strcmp (status, "available") == 0)
+                status = "best available";
+        else if (strcmp (status, "found") == 0)
+                status = "not found";
+        else if (strcmp (status, "supported") == 0)
+                status = "not supported";
+        else if (strcmp (status, "correct") == 0)
+                status = "good";
+        else if (strcmp (status, "incorrect") == 0)
+                status = "bad";
+
+        if (out_name != NULL)
+                *out_name = g_strdup (name);
+
+        if (out_status != NULL)
+                *out_status = g_strdup (status);
+
+exit:
+        g_free (copy);
+
+        return (name != NULL && status != NULL);
+}
+
 /**
  * gva_mame_run_game:
  * @name: the name of the game to run
@@ -445,6 +656,22 @@ gva_mame_run_game (const gchar *name,
                         g_string_append (arguments, "-window ");
         }
 
+#ifdef WITH_WNCK
+        if (gva_mame_supports_maximize ())
+                g_string_append (arguments, "-nomaximize ");
+#endif
+
+        if (gva_mame_supports_sound ())
+        {
+                GvaMuteButton *mute_button;
+
+                mute_button = GVA_MUTE_BUTTON (GVA_WIDGET_MAIN_MUTE_BUTTON);
+                if (gva_mute_button_get_muted (mute_button))
+                        g_string_append (arguments, "-nosound ");
+                else
+                        g_string_append (arguments, "-sound ");
+        }
+
         g_string_append_printf (arguments, "%s", name);
 
         /* Execute the command "${mame} ${name}". */
@@ -506,6 +733,22 @@ gva_mame_record_game (const gchar *name,
                         g_string_append (arguments, "-window ");
         }
 
+#ifdef WITH_WNCK
+        if (gva_mame_supports_maximize ())
+                g_string_append (arguments, "-nomaximize ");
+#endif
+
+        if (gva_mame_supports_sound ())
+        {
+                GvaMuteButton *mute_button;
+
+                mute_button = GVA_MUTE_BUTTON (GVA_WIDGET_MAIN_MUTE_BUTTON);
+                if (gva_mute_button_get_muted (mute_button))
+                        g_string_append (arguments, "-nosound ");
+                else
+                        g_string_append (arguments, "-sound ");
+        }
+
         g_string_append_printf (arguments, "-record %s %s", inpname, name);
 
         /* Execute the command "${mame} -record ${inpname} ${name}". */
@@ -564,6 +807,22 @@ gva_mame_playback_game (const gchar *name,
                         g_string_append (arguments, "-window ");
         }
 
+#ifdef WITH_WNCK
+        if (gva_mame_supports_maximize ())
+                g_string_append (arguments, "-nomaximize ");
+#endif
+
+        if (gva_mame_supports_sound ())
+        {
+                GvaMuteButton *mute_button;
+
+                mute_button = GVA_MUTE_BUTTON (GVA_WIDGET_MAIN_MUTE_BUTTON);
+                if (gva_mute_button_get_muted (mute_button))
+                        g_string_append (arguments, "-nosound ");
+                else
+                        g_string_append (arguments, "-sound ");
+        }
+
         g_string_append_printf (arguments, "%s -playback %s", name, inpname);
 
         /* Execute the command "${mame} ${name} -playback ${inpname}". */
@@ -594,21 +853,21 @@ gva_mame_playback_game (const gchar *name,
 gchar *
 gva_mame_get_save_state_file (const gchar *name)
 {
-        gchar *directory;
+        const gchar *directory;
         gchar *basename;
         gchar *filename = NULL;
         GError *error = NULL;
 
-        directory = gva_mame_get_config_value ("state_directory", &error);
+        directory = gva_mame_get_state_directory (&error);
         gva_error_handle (&error);
         if (directory == NULL)
-                goto exit;
+                return NULL;
 
         /* As of 0.115, MAME stores save state files as "name/auto.sta". */
         filename = g_build_filename (directory, name, "auto.sta", NULL);
 
         if (g_file_test (filename, G_FILE_TEST_IS_REGULAR))
-                goto exit;
+                return filename;
         else
                 g_free (filename);
 
@@ -618,16 +877,11 @@ gva_mame_get_save_state_file (const gchar *name)
         g_free (basename);
 
         if (g_file_test (filename, G_FILE_TEST_IS_REGULAR))
-                goto exit;
+                return filename;
         else
                 g_free (filename);
 
-        filename = NULL;  /* file not found */
-
-exit:
-        g_free (directory);
-
-        return filename;
+        return NULL;  /* file not found */
 }
 
 /**
@@ -651,10 +905,88 @@ gva_mame_delete_save_state (const gchar *name)
         g_free (filename);
 }
 
+/**
+ * gva_mame_get_input_directory:
+ * @error: return location for a #GError, or %NULL
+ *
+ * Returns the value of the "input_directory" configuration key.  The
+ * function caches the result to avoid excessive MAME invocations.  If
+ * an error occurs, the function returns %NULL and sets @error.
+ *
+ * Returns: the directory containing MAME input files, or %NULL
+ **/
+const gchar *
+gva_mame_get_input_directory (GError **error)
+{
+        static gchar *directory = NULL;
+
+        if (G_UNLIKELY (directory == NULL)) {
+                const gchar *config_key = "input_directory";
+                directory = gva_mame_get_config_value (config_key, error);
+        }
+
+        return directory;
+}
+
+/**
+ * gva_mame_get_snapshot_directory:
+ * @error: return location for a #GError, or %NULL
+ *
+ * Returns the value of the "snapshot_directory" configuration key.  The
+ * function caches the result to avoid excessive MAME invocations.  If an
+ * error occurs, the function returns %NULL and sets @error.
+ *
+ * Returns: the directory containing MAME snapshot files, or %NULL
+ **/
+const gchar *
+gva_mame_get_snapshot_directory (GError **error)
+{
+        static gchar *directory = NULL;
+
+        if (G_UNLIKELY (directory == NULL)) {
+                const gchar *config_key = "snapshot_directory";
+                directory = gva_mame_get_config_value (config_key, error);
+        }
+
+        return directory;
+}
+
+/**
+ * gva_mame_get_state_directory:
+ * @error: return location for a #GError, or %NULL
+ *
+ * Returns the value of the "state_directory" configuration key.  The
+ * function caches the result to avoid excessive MAME invocations.  If
+ * an error occurs, the function returns %NULL and sets @error.
+ *
+ * Returns: the directory containing MAME save state files, or %NULL
+ **/
+const gchar *
+gva_mame_get_state_directory (GError **error)
+{
+        static gchar *directory = NULL;
+
+        if (G_UNLIKELY (directory == NULL)) {
+                const gchar *config_key = "state_directory";
+                directory = gva_mame_get_config_value (config_key, error);
+        }
+
+        return directory;
+}
+
 /*****************************************************************************
  * Documentation for the rest of the public MAME interface
  *****************************************************************************/
 
+/**
+ * gva_mame_get_path_sep:
+ *
+ * Returns the search path delimiter string.
+ * XMAME uses UNIX-style ':' whereas SDLMAME uses Window-style ';'.
+ *
+ * Returns: search path delimiter string
+ **/
+
 /**
  * gva_mame_get_version:
  * @error: return location for a #GError, or %NULL
